#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import

import os
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file
import torch
import torchvision.models as models

class pytorch_extractor(base_extractor):

    architecture_map = sorted(name for name in models.__dict__
        if name.islower() and not name.startswith("__")
        and callable(models.__dict__[name]))


    @classmethod
    def help(cls):
        print ('Support frameworks: {}'.format(cls.architecture_map))


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            architecture_file = path + "imagenet_{}.pth".format(architecture)
            if not os.path.exists(architecture_file):
                kwargs = {}
                if architecture == 'inception_v3':
                    kwargs['transform_input'] = False
                model = models.__dict__[architecture](pretrained=True, **kwargs)
                torch.save(model, architecture_file)
                print("PyTorch pretrained model is saved as [{}].".format(architecture_file))
            else:
                print("File [{}] existed!".format(architecture_file))

            return architecture_file

        else:
            return None


    @classmethod
    def inference(cls, architecture, path, image_path):
        model = torch.load(path)

        model.eval()

        import numpy as np
        func = TestKit.preprocess_func['pytorch'][architecture]
        img = func(image_path)
        img = np.transpose(img, (2, 0, 1))

        img = np.expand_dims(img, 0).copy()

        data = torch.from_numpy(img)
        data = torch.autograd.Variable(data, requires_grad=False)

        predict = model(data).data.numpy()
        predict = np.squeeze(predict)

        return predict



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import numpy as np
import sys
import os
import tensorflow as tf
from mmdnn.conversion.examples.imagenet_test import TestKit

class TestTF(TestKit):

    def __init__(self):
        super(TestTF, self).__init__()

        self.truth['mxnet']['resnet152-11k'] = [(1278, 0.49070787), (1277, 0.21392652), (282, 0.12979421), (1282, 0.066355646), (1224, 0.022040566)]

        self.input, self.model = self.MainModel.KitModel(self.args.w)
        # self.input, self.model, self.testop = self.MainModel.KitModel(self.args.w)


    def preprocess(self, image_path):
        x = super(TestTF, self).preprocess(image_path)
        self.data = np.expand_dims(x, 0)


    def print_result(self):
        with tf.Session() as sess:
            init = tf.global_variables_initializer()
            sess.run(init)
            predict = sess.run(self.model, feed_dict = {self.input : self.data})

        super(TestTF, self).print_result(predict)


    def print_intermediate_result(self, layer_name, if_transpose = False):
        # testop = tf.get_default_graph().get_operation_by_name(layer_name)
        testop = self.testop
        with tf.Session() as sess:
            init = tf.global_variables_initializer()
            sess.run(init)
            intermediate_output = sess.run(testop, feed_dict = {self.input : self.data})

        super(TestTF, self).print_intermediate_result(intermediate_output, if_transpose)


    def inference(self, image_path):
        self.preprocess(image_path)

        # self.print_intermediate_result('conv1_7x7_s2_1', True)

        self.print_result()

        self.test_truth()


    def dump(self, path = None):
        dump_tag = self.args.dump_tag
        if dump_tag == 'SERVING':
            tag_list = [tf.saved_model.tag_constants.SERVING]
        else:
            tag_list = [tf.saved_model.tag_constants.TRAINING]

        if path is None: path = self.args.dump
        with tf.Session() as sess:
            sess.run(tf.global_variables_initializer())

            builder = tf.saved_model.builder.SavedModelBuilder(path)

            tensor_info_input = tf.saved_model.utils.build_tensor_info(self.input)
            tensor_info_output = tf.saved_model.utils.build_tensor_info(self.model)

            prediction_signature = (
                tf.saved_model.signature_def_utils.build_signature_def(
                    inputs={'input': tensor_info_input},
                    outputs={'output': tensor_info_output},
                    method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME
                )
            )

            builder.add_meta_graph_and_variables(
                sess,
                tag_list,
                signature_def_map={
                    tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: prediction_signature
                }
            )

            save_path = builder.save()

            print ('Tensorflow file is saved as [{}], generated by [{}.py] and [{}].'.format(
                save_path, self.args.n, self.args.w))


if __name__=='__main__':
    tester = TestTF()
    if tester.args.dump:
        if tester.args.dump_tag:
            tester.dump()
        else:
            raise ValueError("Need to provide the model type of Tensorflow model.")
    else:
        tester.inference(tester.args.image)



import paddle.v2 as paddle

__all__ = ['resnet_imagenet', 'resnet_cifar10']


def conv_bn_layer(input,
                  ch_out,
                  filter_size,
                  stride,
                  padding,
                  active_type=paddle.activation.Relu(),
                  ch_in=None):
    tmp = paddle.layer.img_conv(
        input=input,
        filter_size=filter_size,
        num_channels=ch_in,
        num_filters=ch_out,
        stride=stride,
        padding=padding,
        act=paddle.activation.Linear(),
        bias_attr=False)
    return paddle.layer.batch_norm(input=tmp, act=active_type)


def shortcut(input, ch_out, stride):
    if input.num_filters != ch_out:
        return conv_bn_layer(input, ch_out, 1, stride, 0,
                             paddle.activation.Linear())
    else:
        return input


def basicblock(input, ch_out, stride):
    short = shortcut(input, ch_out, stride)
    conv1 = conv_bn_layer(input, ch_out, 3, stride, 1)
    conv2 = conv_bn_layer(conv1, ch_out, 3, 1, 1, paddle.activation.Linear())
    return paddle.layer.addto(
        input=[short, conv2], act=paddle.activation.Relu())


def bottleneck(input, ch_out, stride):
    short = shortcut(input, ch_out * 4, stride)
    conv1 = conv_bn_layer(input, ch_out, 1, stride, 0)
    conv2 = conv_bn_layer(conv1, ch_out, 3, 1, 1)
    conv3 = conv_bn_layer(conv2, ch_out * 4, 1, 1, 0,
                          paddle.activation.Linear())
    return paddle.layer.addto(
        input=[short, conv3], act=paddle.activation.Relu())


def layer_warp(block_func, input, ch_out, count, stride):
    conv = block_func(input, ch_out, stride)
    for i in range(1, count):
        conv = block_func(conv, ch_out, 1)
    return conv


def resnet_imagenet(input, class_dim, depth=50):
    cfg = {
        18: ([2, 2, 2, 1], basicblock),
        34: ([3, 4, 6, 3], basicblock),
        50: ([3, 4, 6, 3], bottleneck),
        101: ([3, 4, 23, 3], bottleneck),
        152: ([3, 8, 36, 3], bottleneck)
    }
    stages, block_func = cfg[depth]
    conv1 = conv_bn_layer(
        input, ch_in=3, ch_out=64, filter_size=7, stride=2, padding=3)
    pool1 = paddle.layer.img_pool(input=conv1, pool_size=3, stride=2)
    res1 = layer_warp(block_func, pool1, 64, stages[0], 1)
    res2 = layer_warp(block_func, res1, 128, stages[1], 2)
    res3 = layer_warp(block_func, res2, 256, stages[2], 2)
    res4 = layer_warp(block_func, res3, 512, stages[3], 2)
    pool2 = paddle.layer.img_pool(
        input=res4, pool_size=7, stride=1, pool_type=paddle.pooling.Avg())
    out = paddle.layer.fc(input=pool2,
                          size=class_dim,
                          act=paddle.activation.Softmax())
    return out


def resnet_cifar10(input, class_dim, depth=32):
    # depth should be one of 20, 32, 44, 56, 110, 1202
    assert (depth - 2) % 6 == 0
    n = (depth - 2) / 6
    nStages = {16, 64, 128}
    conv1 = conv_bn_layer(
        input, ch_in=3, ch_out=16, filter_size=3, stride=1, padding=1)
    res1 = layer_warp(basicblock, conv1, 16, n, 1)
    res2 = layer_warp(basicblock, res1, 32, n, 2)
    res3 = layer_warp(basicblock, res2, 64, n, 2)
    pool = paddle.layer.img_pool(
        input=res3, pool_size=8, stride=1, pool_type=paddle.pooling.Avg())
    out = paddle.layer.fc(input=pool,
                          size=class_dim,
                          act=paddle.activation.Softmax())
    return out



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
import os
import keras
from keras import backend as K
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class keras_extractor(base_extractor):

    MMDNN_BASE_URL = 'http://mmdnn.eastasia.cloudapp.azure.com:89/models/'

    architecture_map = {
        'inception_v3'        : lambda : keras.applications.inception_v3.InceptionV3(input_shape=(299, 299, 3)),
        'vgg16'               : lambda : keras.applications.vgg16.VGG16(),
        'vgg19'               : lambda : keras.applications.vgg19.VGG19(),
        'resnet50'            : lambda : keras.applications.resnet50.ResNet50(),
        'mobilenet'           : lambda : keras.applications.mobilenet.MobileNet(),
        'xception'            : lambda : keras.applications.xception.Xception(input_shape=(299, 299, 3)),
        'inception_resnet_v2' : lambda : keras.applications.inception_resnet_v2.InceptionResNetV2(input_shape=(299, 299, 3)),
        'densenet'            : lambda : keras.applications.densenet.DenseNet201(),
        'nasnet'              : lambda : keras.applications.nasnet.NASNetLarge()
    }

    thirdparty_map = {
        'yolo2'    : MMDNN_BASE_URL + 'keras/yolo2.h5',
    }

    image_size = {
        'inception_v3'      : 299,
        'vgg16'             : 224,
        'vgg19'             : 224,
        'resnet'            : 224,
        'mobilenet'         : 224,
        'xception'          : 299,
        'inception_resnet'  : 299,
        'densenet'          : 224,
        'nasnet'            : 331,
    }

    @classmethod
    def help(cls):
        print ('Support frameworks: {}'.format(set().union(cls.architecture_map.keys(), cls.thirdparty_map.keys())))


    @classmethod
    def download(cls, architecture, path="./"):
        if architecture in cls.thirdparty_map:
            weight_file = download_file(cls.thirdparty_map[architecture], directory=path)
            return weight_file

        elif cls.sanity_check(architecture):
            output_filename = path + 'imagenet_{}.h5'.format(architecture)
            if os.path.exists(output_filename) == False:
                model = cls.architecture_map[architecture]()
                model.save(output_filename)
                print("Keras model {} is saved in [{}]".format(architecture, output_filename))
                K.clear_session()
                del model
                return output_filename

            else:
                print("File [{}] existed, skip download.".format(output_filename))
                return output_filename

        else:
            return None


    @classmethod
    def inference(cls, architecture, files, path, image_path):
        if architecture in cls.thirdparty_map:
            model = keras.models.load_model(files)

        elif cls.sanity_check(architecture):
            model = cls.architecture_map[architecture]()

        else:
            model = None

        if model:
            import numpy as np
            func = TestKit.preprocess_func['keras'][architecture]
            img = func(image_path)
            img = np.expand_dims(img, axis=0)
            predict = model.predict(img)
            predict = np.squeeze(predict)
            K.clear_session()
            del model
            return predict

        else:
            return None



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------
import os
import paddle.v2 as paddle
import paddle.trainer_config_helpers.layers as layers
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph


class PaddleGraphNode(GraphNode):

    def __init__(self, layer):
        super(PaddleGraphNode, self).__init__(layer)


    @property
    def name(self):
        return self.layer.name


    @property
    def type(self):
        return self.layer.type


    @property
    def paddle_layer(self):
        return self.layer



class PaddleGraph(Graph):

    def __init__(self, model):
        from paddle.proto import ModelConfig_pb2
        # sanity check.
        if not isinstance(model, ModelConfig_pb2.ModelConfig):
            raise TypeError("PaddlePaddle layer of type %s is not supported." % type(model))
        super(PaddleGraph, self).__init__(model)
        self.model = model


    def build(self):
        self.input_layers = list()
        for layer in self.model.layers:
            self.layer_map[layer.name] = PaddleGraphNode(layer)
            self.layer_name_map[layer.name] = layer.name

            for input_layer in layer.inputs:
                self._make_connection(input_layer.input_layer_name, layer.name)

        super(PaddleGraph, self).build()


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import numpy as np
import tensorflow
from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2
from mmdnn.conversion.tensorflow.tensorflow_graph import TensorflowGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.DataStructure.parser import Parser
from tensorflow.tools.graph_transforms import TransformGraph
from mmdnn.conversion.rewriter.utils import *


class TensorflowParser(Parser):

    skip_prefix = [
        "^",
        "train_op",
        "save",
        "gradients",
        "global_step",
        "distort_image",
        "Adagrad",
    ]

    skip_scope = [
        "random_uniform",
        "Initializer",
        "optimizer",
        "weight_loss",
        "parallel_read",
        "case"
    ]

    skip_type = set([
        "L2Loss",
        "VariableV2",
        "Const",
        "Assign",
        "RandomUniform",
        "FIFOQueueV2"
    ])

    dtype_map = {
        0  : graph_pb2.DT_UNDEFINED,
        1  : graph_pb2.DT_FLOAT32,
        2  : graph_pb2.DT_FLOAT64,
        3  : graph_pb2.DT_INT32,
        4  : graph_pb2.DT_UINT8,
        5  : graph_pb2.DT_INT16,
        6  : graph_pb2.DT_INT8,
        7  : graph_pb2.DT_STRING,
        9  : graph_pb2.DT_INT64,
        10 : graph_pb2.DT_BOOL,
        19 : graph_pb2.DT_FLOAT16
    }


    @property
    def src_graph(self):
        return self.tf_graph


    @staticmethod
    def _shapeToStr(shapes):
        return [dim.size if dim.size > 0 else 1 for dim in shapes.dim]


    @staticmethod
    def _load_meta(model_network_path):
        """Load a tensorflow meta file from disk

        Parameters
        ----------
        model_network_path: str
            Path where the model network path is (protobuf meta file)

        Returns
        -------
        model: A tensorflow protobuf file
        """
        from tensorflow.core.protobuf import meta_graph_pb2
        from mmdnn.conversion.common.IR.IR_graph import load_protobuf_from_file

        meta_graph = meta_graph_pb2.MetaGraphDef()
        load_protobuf_from_file(meta_graph, model_network_path)
        graph = meta_graph.graph_def

        print ("Tensorflow model file [%s] loaded successfully." % model_network_path)
        return graph


    @staticmethod
    def _load_weights(model_weight_path):
        """Load a tensorflow checkpoint file from disk

        Parameters
        ----------
        model_weight_path: str
            Path where the weight path is (checkpoint file)

        Returns
        -------
        model: tensor name --> ndarry
        """
        reader = tensorflow.train.NewCheckpointReader(model_weight_path)
        var_to_shape_map = reader.get_variable_to_shape_map()
        data = dict()
        for name in var_to_shape_map:
            tensor = reader.get_tensor(name)
            data[name] = tensor

        print ("Tensorflow checkpoint file [%s] loaded successfully. [%d] variables loaded." % (model_weight_path, len(data)))
        return data


    @staticmethod
    def _get_scopes(layer_name):
        return layer_name.split('/')


    def check_const(self, node):
        while node:
            if node.type == "Const":
                return node
            elif node.type == "NoOp":
                return None
            else:
                node =  self.get_parent(node.name, [0])

    def _add_constant_node(self, source_node):
        parent_ids=range(len(source_node.in_edges))
        for idx in parent_ids:
            parent_node = self.tf_graph.get_node(source_node.in_edges[idx])
            if parent_node.type == 'Const':
                self._rename_Const(parent_node)
    
    def _rename_Const(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count=0, new_op='Constant') # Constant
        value = source_node.get_attr('value')
        if value.float_val:
            shape = tuple(self.tensor_shape_to_list(value.tensor_shape))
            value = np.full(shape, value.float_val[0])
        elif value.int_val:
            shape = tuple(self.tensor_shape_to_list(value.tensor_shape))
            value = np.full(shape, value.int_val[0])
        else:
            value = np.array(tensor_util.MakeNdarray(value).tolist())
        
        if value.ndim > 1:
            self.set_weight(source_node.name, 'value', value)
        else:
            kwargs = {'value': value}
            assign_IRnode_values(IR_node, kwargs)


    def _convert_reduction_operators(self, source_node, new_op = None):
        IR_node = self._convert_identity_operation(source_node, 0, 1, new_op)

        # keep dims
        IR_node.attr['keepdims'].b = source_node.layer.attr['keep_dims'].b

        # axes
        axes = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        axes = tensor_util.MakeNdarray(axes)
        IR_node.attr['axes'].list.i.extend(axes)


    def _convert_layers_batchnorm(self, source_node):
        # name, op
        IR_node = self.IR_graph.node.add()
        TensorflowParser._copy_and_reop(source_node, IR_node, 'BatchNorm')

        # epsilon
        epsilon = self.get_parent(source_node.name, [1])
        IR_node.attr['epsilon'].f = epsilon.layer.attr['value'].tensor.float_val[0]

        # moving variance (var)
        moving_variance = self.get_parent(source_node.name, [0, 0])
        # print(moving_variance.name)
        if self.weight_loaded and moving_variance.name in self.ckpt_data.keys():
            self.set_weight(source_node.name, 'var', self.ckpt_data[moving_variance.name])

        # gamma (scale)
        gamma = self.get_son(source_node.name, [0, 0], True)
        gamma = self.get_parent(gamma.name, [1, 0], True)
        if gamma is None or not gamma.type.startswith('Variable'):
            IR_node.attr['scale'].b = False
            output_node = self.get_son(source_node.name, [0, 0, 0], True)
        else:
            IR_node.attr['scale'].b = True
            if self.weight_loaded:
                self.set_weight(source_node.name, 'scale', self.ckpt_data[gamma.name])
            output_node = self.get_son(source_node.name, [0, 0, 0, 0], True)
            if output_node.type == 'Sub':
                output_node = self.get_son(source_node.name, [0, 0, 1, 0], True)

        # mean
        mean = self.get_parent(output_node.name, [1, 1, 0, 0], True)
        if self.weight_loaded and mean.name in self.ckpt_data.keys():
            self.set_weight(source_node.name, 'mean', self.ckpt_data[mean.name])

        # bias
        bias = self.get_parent(output_node.name, [1, 0, 0], True)
        if bias is None or not bias.type.startswith('Variable'):
            IR_node.attr['bias'].b = False
        else:
            IR_node.attr['bias'].b = True
            if self.weight_loaded:
                self.set_weight(source_node.name, 'bias', self.ckpt_data[bias.name])

        # input node
        assert output_node.type == 'Add'
        input_node = self.get_parent(output_node.name, [0, 0])
        IR_node.input.append(input_node.real_name)

        # output node
        output_node.real_name = source_node.name


    def __init__(self, meta_file, checkpoint_file, dest_nodes, inputShape = None, in_nodes = None):
        super(TensorflowParser, self).__init__()

        # load model files into TensorFlow graph
        if meta_file:
            model = TensorflowParser._load_meta(meta_file)

        if checkpoint_file:
            self.ckpt_data = TensorflowParser._load_weights(checkpoint_file)
            self.weight_loaded = True

        # extract subgraph using in_nodes and dest_nodes
        if in_nodes != None and inputShape != None:
            from tensorflow.python.tools import strip_unused_lib
            from tensorflow.python.framework import dtypes
            from tensorflow.python.platform import gfile
            model = strip_unused_lib.strip_unused(
                    input_graph_def = model,
                    input_node_names = in_nodes,
                    output_node_names = dest_nodes,
                    placeholder_type_enum = dtypes.float32.as_datatype_enum)

            input_list = [None]
            for i in range(len(inputShape)):
                input_list.append(tensorflow.Dimension(inputShape[i]))
            tensor_input = tensorflow.TensorShape(input_list)
            # Build network graph
            self.tf_graph = TensorflowGraph(model)
            for node in self.tf_graph.model.node:
                if node.name in in_nodes:
                    node.attr['shape'].shape.CopyFrom(tensor_input.as_proto())
                    node.attr['_output_shapes'].list.shape.pop()  #unknown_rank pop
                    node.attr['_output_shapes'].list.shape.extend([tensor_input.as_proto()])

        # extract subgraph using dest_nodes
        elif dest_nodes != None:
            from tensorflow.python.framework.graph_util import extract_sub_graph
            model = extract_sub_graph(model, dest_nodes)
            self.tf_graph = TensorflowGraph(model)

        else:
            self.tf_graph = TensorflowGraph(model)


        # Graph Transform
        transforms = ["fold_constants(ignore_errors=true)"]

        #  Get input node name
        if not in_nodes:
            in_nodes = []
            for node in model.node:
                if node.op == 'Placeholder':
                    in_nodes.append(node.name)

        transformed_graph_def = TransformGraph(model, in_nodes,
                                            dest_nodes, transforms)
        in_type_list = {}
        in_shape_list = {}

        for n in transformed_graph_def.node:
            if n.name in in_nodes:
                in_type_list[n.name] = n.attr['dtype'].type
                in_node_shape = n.attr['shape'].shape
                in_node_shape_str = self._shapeToStr(in_node_shape)
                in_shape_list[n.name] = in_node_shape_str

        dtype = tensorflow.float32
        with tensorflow.Graph().as_default() as g:
            input_map = {}
            for in_node in in_nodes:
                if in_type_list[in_node] == 1 or in_type_list[in_node] == 0:
                    dtype = tensorflow.float32

                elif in_type_list[in_node] == 3:
                    dtype = tensorflow.int32

                elif in_type_list[in_node] == 10:
                    dtype = tensorflow.bool
                
                x = tensorflow.placeholder(dtype, shape = in_shape_list[in_node])
                input_map[in_node] = x

            tensorflow.import_graph_def(transformed_graph_def, name='', input_map=input_map)

        with tensorflow.Session(graph = g) as sess:

            meta_graph_def = tensorflow.train.export_meta_graph(filename='./my-model.meta')
            model = meta_graph_def.graph_def

        self.tf_graph = TensorflowGraph(model)
        self.tf_graph.build()

        process_graph(self.tf_graph, self.ckpt_data)

    @classmethod
    def _skip_node(cls, source_node):
        if source_node.covered:
            return True

        for prefix in cls.skip_prefix:
            if source_node.name.startswith(prefix):
                return True

        scopes = TensorflowParser._get_scopes(source_node.name)

        for s in scopes:
            if s in cls.skip_scope:
                return True

        return False

    @staticmethod
    def tensor_shape_to_list(shapes):
        if isinstance(shapes, attr_value_pb2.AttrValue):
            return [dim.size for dim in shapes.shape.dim]
        elif isinstance(shapes, attr_value_pb2.tensorflow_dot_core_dot_framework_dot_tensor__shape__pb2.TensorShapeProto):
            return [dim.size for dim in shapes.dim]
        else:
            ret = []
            for shape in shapes:
                this_one = [dim.size for dim in shape.dim]
                ret.append(this_one)
            return ret

    '''
    check current source_node wether has input weights. If it has, set the weights into weight dict and remove the input edge.
    return edges' index which do not include edge connecting weights 
    '''
    def _check_weights(self, source_node, start_edge_id = 0, in_edge_count = None):
        if in_edge_count == None: in_edge_count = len(source_node.in_edges) - start_edge_id
        valid_pre_ids = []

        for pre_idx in range(start_edge_id, start_edge_id + in_edge_count):
            pre_node = self.get_parent(source_node.name, [pre_idx])
            if pre_node.type == 'Identity' and pre_node.name.split('/')[-1] == 'read':
                weight_node = self.get_parent(pre_node.name, [0])
                assert 'Variable' in weight_node.type
                self.set_weight(source_node.name, 'weights', self.ckpt_data[weight_node.name])
                source_node.feed_weights = True
            else:
                valid_pre_ids.append(pre_idx)

        return valid_pre_ids

    def _convert_padding(self, source_node, IR_node, kernel_size):
        # TODO: Fused conv and pool with padding is different from defused operators
        input_node = self.get_parent(source_node.name, [0])
        input_shape = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        if source_node.get_attr('padding') == 'VALID':
            dims = len(input_shape)
            assign_IRnode_values(IR_node, {'auto_pad' : "VALID", 'pads' : [0, 0] * dims})

        elif source_node.get_attr('padding') == 'SAME':
            padding = compute_tf_same_padding(
                input_shape,
                kernel_size,
                source_node.get_attr('strides'))
            assign_IRnode_values(IR_node, {'auto_pad' : "SAME_UPPER", 'pads' : padding})

        else:
            assert False


    def _convert_pooling(self, source_node, pool_type):
        IR_node = self._convert_identity_operation(source_node, new_op='Pool')
        kwargs = {}

        # strides
        kwargs['strides'] = source_node.get_attr('strides')

        # window_shape
        kwargs['kernel_shape'] = source_node.get_attr('ksize')

        # pool type
        kwargs['pooling_type'] = pool_type

        # padding
        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][1:-1])

        assign_IRnode_values(IR_node, kwargs)


    def gen_IR(self):
        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)

            if self._skip_node(current_node):
                continue

            node_type = current_node.type

            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)
            else:
                self.rename_UNKNOWN(current_node)


    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        if new_op == None: new_op = source_node.type
        IR_node.name = source_node.name
        IR_node.op = new_op

        kwargs = {}
        if 'data_format' in source_node.layer.attr:
            kwargs['data_format'] = source_node.get_attr('data_format')

        if 'dtype' in source_node.layer.attr:
            assert source_node.layer.attr['dtype'].type in TensorflowParser.dtype_map, 'type [{}] is unknown.'.format(source_node.layer.attr['dtype'].type)
            IR_node.attr["dtype"].type = TensorflowParser.dtype_map[source_node.layer.attr['dtype'].type]

        if '_output_shapes' in source_node.layer.attr:
            IR_node.attr["_output_shapes"].MergeFromString(source_node.layer.attr['_output_shapes'].SerializeToString())

        if hasattr(source_node, 'feed_weights'):
            kwargs["feed_weights"] = True

        if hasattr(source_node, 'kwargs'):
            kwargs.update(source_node.kwargs)

        kwargs['scope'] = source_node.scope

        assign_IRnode_values(IR_node, kwargs)


    def _convert_inedge(self, source_node, IR_node, start_idx = 0, end_idx = None, in_ids=None):
        if end_idx == None: end_idx = len(source_node.in_edges) - start_idx
        if not in_ids:
            in_ids = range(start_idx, end_idx + start_idx)

        for idx in in_ids:
            if ':' in source_node.in_edges[idx]:
                input_tensor = self.src_graph.get_node(source_node.in_edges[idx]).real_name + ':' + source_node.in_edges[idx].split(':')[1]
            else:
                input_tensor = self.src_graph.get_node(source_node.in_edges[idx]).real_name

            IR_node.input.append(input_tensor)




    def _get_bias(self, source_node, IR_node):
        if not source_node.out_edges:
            return

        add_node = self.tf_graph.get_node(source_node.out_edges[0])
        if add_node.type != "Add" and add_node.type != "BiasAdd":
            return

        variable = self.tf_graph.get_node(add_node.in_edges[1])
        if variable.type != "Identity":
            return
        variable = self.tf_graph.get_node(variable.in_edges[0])

        assert variable.layer.attr['shape'].shape.dim[0].size == IR_node.attr['kernel_shape'].list.i[-1]

        if self.weight_loaded:
            assert variable.name in self.ckpt_data
            current_layer = self.weights[source_node.name]
            current_layer['bias'] = self.ckpt_data[variable.name]

        add_node.real_name = IR_node.name
        add_node.covered = True
        IR_node.attr['use_bias'].b = True


    @staticmethod
    def _copy_shape(source_node, IR_node):
        assert 'shape' in source_node.layer.attr
        if source_node.layer.attr['shape'].list.shape:
            IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['shape'].list.shape[0].SerializeToString())
        else:
            IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['shape'].shape.SerializeToString())


    def rename_UNKNOWN(self, source_node):
        if source_node.type in self.skip_type:
            return
        print("TensorflowEmitter has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))
        return


    def rename_Placeholder(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='DataInput')
        # shape
        TensorflowParser._copy_shape(source_node, IR_node)
        IR_node.attr['shape'].shape.dim[0].size = -1
        IR_node.attr['_output_shapes'].list.shape[0].dim[0].size = -1


    def rename_Conv2D(self, source_node):
        """
        weights: name_weights, name_bias
        """
        IR_node = self._convert_identity_operation(source_node, 0, 1, 'Conv')

        kwargs = {}

        # strides
        kwargs['strides'] = source_node.get_attr('strides')

        # input[1] : W
        # filter
        W = self.tf_graph.get_node(source_node.layer.input[1])
        if W.type == 'Const':
            kwargs['kernel_shape'] = tensor_shape = self.tensor_shape_to_list(W.layer.attr['value'].tensor.tensor_shape)
        else:
            W = self.tf_graph.get_node(W.layer.input[0]).layer
            kwargs['kernel_shape'] = self.tensor_shape_to_list(W.attr['shape'])

        # padding
        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][:-2])

        if self.weight_loaded:
             self.set_weight(source_node.name, 'weights', self.ckpt_data[W.name])

        assign_IRnode_values(IR_node, kwargs)
        # output[0] : B
        self._get_bias(source_node, IR_node)


    def _convert_identity_operation(self, source_node, start_edge_id = 0, in_edge_count = None, new_op = None):
        IR_node = self.IR_graph.node.add()
        in_ids = self._check_weights(source_node, start_edge_id, in_edge_count)
        TensorflowParser._copy_and_reop(source_node, IR_node, new_op)
        self._convert_inedge(source_node, IR_node, start_edge_id, in_edge_count, in_ids)
        return IR_node


    def rename_Relu(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Softmax(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Relu6(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Add(self, source_node):
        if not source_node.covered:
            scopes = self._get_scopes(source_node.name)
            if len(scopes) < 3:
                self._convert_identity_operation(source_node)

            elif scopes[-2] == 'dropout':
                # converted [dropout]
                pass

            elif scopes[-2] == 'batchnorm':
                # convert [tf.contrib.layers.batch_norm]
                self._convert_layers_batchnorm(source_node)

            else:
                # normal Add
                self._add_constant_node(source_node)
                self._convert_identity_operation(source_node)


    def rename_Sub(self, source_node):
        self._add_constant_node(source_node)
        self._convert_identity_operation(source_node)


    def rename_Reshape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 1)
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        assign_IRnode_values(IR_node, kwargs)


    def rename_Abs(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 1, new_op = 'Abs')


    def rename_Square(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 1, new_op = 'Square')


    def rename_MatMul(self, source_node):

        W = self.tf_graph.get_node(self.tf_graph.get_node(source_node.in_edges[1]).in_edges[0])

        if 'Variable' in W.type:

            """
            weights: name_weights, name_bias
            """
            IR_node = self._convert_identity_operation(source_node, in_edge_count = 1)

            # units
            units = source_node.layer.attr['_output_shapes'].list.shape[-1].dim[-1].size
            IR_node.attr['units'].i = units

            # Weights
            W = self.tf_graph.get_node(self.tf_graph.get_node(source_node.in_edges[1]).in_edges[0])
            if self.weight_loaded:
                self.set_weight(source_node.name, 'weights', self.ckpt_data[W.name])

            if source_node.out_edges and (self.tf_graph.get_node(source_node.out_edges[0]).type == 'Add' or self.tf_graph.get_node(source_node.out_edges[0]).type == 'BiasAdd'):
                add_node = self.tf_graph.get_node(source_node.out_edges[0])
                add_node.covered = True
                add_node.real_name = source_node.real_name
                # FullyConnected Layer
                # name, op
                TensorflowParser._copy_and_reop(source_node, IR_node, 'FullyConnected')

                # get Bias
                B = self.tf_graph.get_node(self.tf_graph.get_node(source_node.out_edges[0]).in_edges[1]).in_edges[0]
                if self.weight_loaded:
                    self.set_weight(source_node.name, 'bias', self.ckpt_data[B])
                IR_node.attr['use_bias'].b = True

            else:
                # Matmul Layer
                TensorflowParser._copy_and_reop(source_node, IR_node, 'FullyConnected')
                assign_IRnode_values(IR_node, {'use_bias' : False})
        else:
            self._convert_identity_operation(source_node, new_op='MatMul')


    def rename_RealDiv(self, source_node):
        scopes = self._get_scopes(source_node.name)

        # Deal Dropout
        if len(scopes) > 1 and scopes[-2][:7] == 'dropout':
            IR_node = self._convert_identity_operation(source_node, in_edge_count = 1, new_op = 'Dropout')

            # keep prob
            if 'value' in self.tf_graph.get_node(source_node.layer.input[1]).layer.attr:
                IR_node.attr['keep_prob'].f = self.tf_graph.get_node(source_node.layer.input[1]).layer.attr['value'].tensor.float_val[0]
            else:
                IR_node.attr['keep_prob'].f = 1.0

            # Remove nodes
            # Mul
            mul_node = self.tf_graph.get_node(source_node.out_edges[0])
            assert mul_node.type == "Mul"
            mul_node.covered = True
            mul_node.real_name = source_node.name

            # Floor
            floor_node = self.tf_graph.get_node(mul_node.in_edges[1])
            assert floor_node.type == "Floor"
            floor_node.covered = True

        else:
            # print (source_node)
            # print (source_node.layer)
            # assert False
            self._convert_identity_operation(source_node, new_op='Div')


    def rename_Floor(self, source_node):
        scopes = self._get_scopes(source_node.name)
        assert scopes[-2] == 'dropout'


    def rename_MaxPool(self, source_node):
        self._convert_pooling(source_node, b'MAX')


    def rename_AvgPool(self, source_node):
        self._convert_pooling(source_node, b'AVG')


    def rename_Identity(self, source_node):
        source_node.real_name =  self.src_graph.get_node(source_node.in_edges[0]).real_name


    def rename_Squeeze(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        IR_node.attr['axes'].MergeFromString(source_node.layer.attr['squeeze_dims'].SerializeToString())

    def rename_QueueDequeueUpToV2(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 0, new_op = 'DataInput')
        IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['_output_shapes'].list.shape[0].SerializeToString())
        IR_node.attr['shape'].shape.dim[0].size = -1
        IR_node.attr['dtype'].type = self.dtype_map[source_node.layer.attr['component_types'].list.type[0]]



    def rename_QueueDequeueManyV2(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 0, new_op = 'DataInput')
        IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['_output_shapes'].list.shape[0].SerializeToString())
        IR_node.attr['shape'].shape.dim[0].size = -1
        IR_node.attr['dtype'].type = self.dtype_map[source_node.layer.attr['component_types'].list.type[0]]

    # def rename_RandomShuffleQueueV2(self, source_node):
    #     # print(source_node.layer)
    #     IR_node = self._convert_identity_operation(source_node, in_edge_count = 0, new_op = 'DataInput')
    #     # IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['_output_shapes'].list.shape[0].SerializeToString())
    #     # IR_node.attr['shape'].shape.dim[0].size = -1
    #     IR_node.attr['dtype'].type = self.dtype_map[source_node.layer.attr['component_types'].list.type[0]]


    def rename_Pad(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count = 1, new_op = 'Pad')
        kwargs = {}
        kwargs['mode'] = 'constant'
        kwargs['constant_values'] = 0.0

        # paddings
        padding = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        shapes = tensor_util.MakeNdarray(padding)
        kwargs['pads'] = convert_tf_pad_to_onnx(shapes)

        assign_IRnode_values(IR_node, kwargs)


    def rename_Mean(self, source_node):
        self._convert_reduction_operators(source_node, new_op = 'ReduceMean')


    def rename_ConcatV2(self, source_node):
        n = len(source_node.in_edges) - 1
        self._add_constant_node(source_node)
        IR_node = self._convert_identity_operation(source_node, in_edge_count = n, new_op = 'Concat')
        axis = self.tf_graph.get_parent(source_node.name, [n])
        IR_node.attr['axis'].i = axis.layer.attr['value'].tensor.int_val[0]


    def rename_DepthwiseConv2dNative(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count=1, new_op='DepthwiseConv')
        kwargs = {}
        kwargs['strides'] = source_node.get_attr('strides')

        input_node = self.src_graph.get_parent(source_node.name, [1])
        kwargs['kernel_shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][:-2])

        if self.weight_loaded:
            weight = self.src_graph.get_parent(source_node.name, [1, 0])
            self.set_weight(source_node.name, 'weights', self.ckpt_data[weight.name])

        assign_IRnode_values(IR_node, kwargs)


    def rename_FusedBatchNorm(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count=1, new_op='BatchNorm')
        IR_node.attr['epsilon'].f = source_node.get_attr('epsilon', 0)

        # gamma (scale)
        scale = self.get_parent(source_node.name, [1], True)

        if scale.type == 'Const':
            value = scale.get_attr('value')
            shape = value.tensor_shape
            assert len(shape.dim) == 1
            shape = shape.dim[0].size

            assert len(value.float_val) == 1
            value = value.float_val[0]

            if np.isclose(value, 1.0):
                IR_node.attr['scale'].b = False
            else:
                IR_node.attr['scale'].b = True
                if self.weight_loaded:
                    self.set_weight(source_node.name, 'scale', np.array([value] * shape))

        else:
            scale = self.get_parent(scale.name, [0], True)
            if self.weight_loaded:
                self.set_weight(source_node.name, 'scale', self.ckpt_data[scale.name])
            IR_node.attr['scale'].b = True

        # bias
        bias = self.get_parent(source_node.name, [2, 0], True)
        IR_node.attr['bias'].b = True

        # Mean
        mean = self.get_parent(source_node.name, [3, 0], True)

        # Var
        var = self.get_parent(source_node.name, [4, 0], True)

        if self.weight_loaded:
            self.set_weight(source_node.name, 'bias', self.ckpt_data[bias.name])
            self.set_weight(source_node.name, 'mean', self.ckpt_data[mean.name])
            self.set_weight(source_node.name, 'var', self.ckpt_data[var.name])


    def rename_Shape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, in_edge_count=1, new_op='Shape')

    def rename_Pack(self, source_node):
        N = len(source_node.layer.input)
        for i in range(N):
            this_node = self.get_parent(source_node.name, [i])
            if this_node.type == 'Const':

                IR_node = self.IR_graph.node.add()
                TensorflowParser._copy_and_reop(this_node, IR_node, 'Constant')
                kwargs = {
                    'value' : this_node.layer.attr['value'].tensor.int_val[0],
                }
                assign_IRnode_values(IR_node, kwargs)

        IR_node = self._convert_identity_operation(source_node, new_op='Pack')
        kwargs = {
            'axis' : source_node.layer.attr['axis'].i,
            'N'    : source_node.layer.attr['N'].i
        }
        assign_IRnode_values(IR_node, kwargs)

    def rename_Gather(self, source_node):

        W = self.src_graph.get_parent(source_node.name, [0])
        W = self.src_graph.get_parent(W.name, [0])

        if 'Variable' in W.type:
            IR_node = self._convert_identity_operation(source_node, new_op='Embedding')

            self.set_weight(source_node.name, "weights", self.ckpt_data[W.name])

            kwargs = {
                'input_dim' : self.ckpt_data[W.name].shape[0],
                'output_dim' : self.ckpt_data[W.name].shape[1],
                'mask_zero' : False
            }
            kwargs['axis'] = 0  # add default
            assign_IRnode_values(IR_node, kwargs)
        else:
            IR_node = self._convert_identity_operation(source_node, new_op='Gather')

        return IR_node

    def rename_GatherV2(self, source_node):
        
        IR_node = self.rename_Gather(source_node)

        kwargs = {}
        kwargs['axis'] = source_node.layer.attr['axis'].i
        assign_IRnode_values(IR_node, kwargs)


    def rename_Transpose(self, source_node):
        IR_node = self._convert_identity_operation(source_node)



    def rename_Sigmoid(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Mul(self, source_node):
        scale1 = self.get_parent(source_node.name, [1], True)
        scale2 = self.get_parent(source_node.name, [0], True)

        if scale1.type == 'Const' or scale2.type == 'Const':
            self._add_constant_node(source_node)
            self._convert_identity_operation(source_node)

        elif scale2.type == 'Identity':
            scale2 = self.get_parent(scale2.name, [0], True)
            assert scale2.type == "VariableV2"
            self.set_weight(source_node.name, 'alpha', self.ckpt_data[scale2.name])
            self._convert_identity_operation(source_node)

        else:
            self._convert_identity_operation(source_node)


    '''
    tf.unpack has been deprecated with replaced tf.unstack
    '''
    def rename_Unpack(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Unstack')
        kwargs = {
            'axis' : source_node.get_attr('axis'),
            'num'  : source_node.get_attr('num')
        }
        assign_IRnode_values(IR_node, kwargs)


    def rename_Split(self, source_node):
        if source_node.get_attr('num_split') == 1:            
            for n in source_node.out_nodes:
                for idx, e in enumerate(n.in_edges):
                    if source_node.name in e:
                        n.in_edges[idx] = e.split(':')[0]

            source_node.real_name = self.get_parent(source_node.name, [1]).real_name

        else:
            IR_node = self._convert_identity_operation(source_node, 1, 1)
            kwargs = {
                'axis' : self.get_parent(source_node.name, [0]).layer.attr['value'].tensor.int_val[0],
                'split' : source_node.get_attr('num_split')
            }
            assign_IRnode_values(IR_node, kwargs)


    def rename_StridedSlice(self, source_node):
        # TODO: Current it is only for slice

        if self.get_parent(source_node.name, [1]).type != 'Const':
            self._add_constant_node(source_node)
            IR_node = self._convert_identity_operation(source_node, new_op='Slice')
            return

        IR_node = self._convert_identity_operation(source_node, in_edge_count=1, new_op='Slice')
        kwargs = {
            'begin_mask' : source_node.get_attr('begin_mask'),
            'end_mask'   : source_node.get_attr('end_mask'),
            'shrink_axis_mask': source_node.get_attr('shrink_axis_mask'),
            'new_axis_mask' :source_node.get_attr('new_axis_mask')
        }

        starts = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        starts = tensor_util.MakeNdarray(starts).tolist()
        kwargs['starts'] = starts

        ends = self.get_parent(source_node.name, [2]).layer.attr['value'].tensor
        ends = tensor_util.MakeNdarray(ends).tolist()
        kwargs['ends'] = ends

        if self.get_parent(source_node.name, [3]) != None:
            strides = self.get_parent(source_node.name, [3]).layer.attr['value'].tensor
            strides = tensor_util.MakeNdarray(strides).tolist()
            kwargs['strides'] = strides

        assign_IRnode_values(IR_node, kwargs)


    def rename_Slice(self, source_node):
        input_node_begin = self.get_parent(source_node.name, [1])
        input_node_size = self.get_parent(source_node.name, [2])

        shape = self.get_parent(source_node.name, [0]).layer.attr['value'].tensor
        shape = tensor_util.MakeNdarray(shape).tolist()

        begin = input_node_begin.get_attr("axis")

        IR_node = self._convert_identity_operation(source_node, in_edge_count=2, new_op='Slice')

        # TODO:  only for 1D
        end = int(input_node_size.layer.attr['value'].tensor.int_val[0]) + begin
        kwargs = {
            'begin_mask' : begin,
            'end_mask' : end
        }

        assign_IRnode_values(IR_node, kwargs)


    def rename_LRN(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        kwargs = {
            "alpha" : source_node.get_attr('alpha') * (source_node.get_attr('depth_radius') * 2 + 1),
            "beta" : source_node.get_attr('beta'),
            "bias" : source_node.get_attr('bias'),
            'size' : source_node.get_attr('depth_radius') + 1
        }
        assign_IRnode_values(IR_node, kwargs)


    def rename_Tanh(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_ExpandDims(self, source_node):
        IR_node = self._convert_identity_operation(source_node, 0, 1, new_op='Unsqueeze')
        
        ax_node = self.get_parent(source_node.name, [1])
        kwargs = {
            'axes': [ax_node.layer.attr['value'].tensor.int_val[0]]
        }
        assign_IRnode_values(IR_node, kwargs)


    def rename_Fill(self, source_node):
        IR_node = self._convert_identity_operation(source_node, 0, 1, new_op='Fill')

        value_node = self.get_parent(source_node.name, [1])
        if value_node.layer.attr['value'].tensor.float_val:
            IR_node.attr['value'].f = value_node.layer.attr['value'].tensor.float_val[0]
        elif value_node.layer.attr['value'].tensor.int_val:
            IR_node.attr['value'].i = value_node.layer.attr['value'].tensor.int_val[0]
        else:
            raise NotImplementedError()


    def rename_Conv2DBackpropInput(self, source_node):
        """
        weights: name_weights, name_bias
        """
        IR_node = self._convert_identity_operation(source_node, new_op = 'ConvTranspose')

        kwargs = {}

        # strides
        kwargs['strides'] = source_node.get_attr('strides')

        # input[1] : W
        # filter
        W = self.tf_graph.get_node(source_node.layer.input[1])
        W = self.tf_graph.get_node(W.layer.input[0]).layer
        kwargs['kernel_shape'] = self.tensor_shape_to_list(W.attr['shape'])

        # padding
        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][:-2])

        if self.weight_loaded:
            self.set_weight(source_node.name, 'weights', self.ckpt_data[W.name])

        assign_IRnode_values(IR_node, kwargs)
        # output[0] : B
        self._get_bias(source_node, IR_node)
    
    def rename_Minimum(self, source_node):
        self._add_constant_node(source_node)
        self._convert_identity_operation(source_node)
    
    def rename_Maxmum(self, source_node):
        self._add_constant_node(source_node)
        self._convert_identity_operation(source_node)


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from six import text_type as _text_type


def generate_label(predict, label_file, offset):
    import os

    if not os.path.exists(label_file):
        return predict

    with open(label_file, 'r') as f:
        labels = [l.rstrip() for l in f]

    ret = []
    for i, j in predict:
        ret.append((labels[i - offset], i, j))

    return ret

def extract_model(args):
    if args.framework == 'caffe':
        from mmdnn.conversion.examples.caffe.extractor import caffe_extractor
        extractor = caffe_extractor()

    elif args.framework == 'keras':
        from mmdnn.conversion.examples.keras.extractor import keras_extractor
        extractor = keras_extractor()

    elif args.framework == 'tensorflow' or args.framework == 'tf':
        from mmdnn.conversion.examples.tensorflow.extractor import tensorflow_extractor
        extractor = tensorflow_extractor()

    elif args.framework == 'mxnet':
        from mmdnn.conversion.examples.mxnet.extractor import mxnet_extractor
        extractor = mxnet_extractor()

    elif args.framework == 'cntk':
        from mmdnn.conversion.examples.cntk.extractor import cntk_extractor
        extractor = cntk_extractor()

    elif args.framework == 'pytorch':
        from mmdnn.conversion.examples.pytorch.extractor import pytorch_extractor
        extractor = pytorch_extractor()

    elif args.framework == 'darknet':
        from mmdnn.conversion.examples.darknet.extractor import darknet_extractor
        extractor = darknet_extractor()

    elif args.framework == 'coreml':
        from mmdnn.conversion.examples.coreml.extractor import coreml_extractor
        extractor = coreml_extractor()

    else:
        raise ValueError("Unknown framework [{}].".format(args.framework))

    files = extractor.download(args.network, args.path)

    if files and args.image:
        predict = extractor.inference(args.network, files, args.path, args.image)
        if type(predict) == list:
            print(predict)

        else:
            if predict.ndim == 1:
                if predict.shape[0] == 1001:
                    offset = 1
                else:
                    offset = 0
                top_indices = predict.argsort()[-5:][::-1]
                predict = [(i, predict[i]) for i in top_indices]
                predict = generate_label(predict, args.label, offset)

                for line in predict:
                    print (line)

            else:
                print (predict.shape)
                print (predict)



def _main():
    import argparse

    parser = argparse.ArgumentParser(description='Extract pre-trained models for frameworks.')

    parser.add_argument(
        '--framework', '-f',
        type=_text_type,
        required=True,
        choices=["caffe", "cntk", "mxnet", "keras", "tensorflow", 'tf', 'pytorch', 'darknet', 'coreml'],
        help="Framework name")

    parser.add_argument(
        '--network', '-n',
        type=_text_type,
        default=None,
        help='Path to the model network file of the external tool (e.g caffe prototxt, keras json')

    parser.add_argument(
        '-i', '--image',
        type=_text_type, help='Test Image Path')

    parser.add_argument(
        '--path', '-p', '-o',
        type=_text_type,
        default='./',
        help='Path to save the pre-trained model files (e.g keras h5)')


    parser.add_argument(
        '-l', '--label',
        type=_text_type,
        default='mmdnn/conversion/examples/data/imagenet_1000.txt',
        help='Path of label.')

    args = parser.parse_args()
    extract_model(args)


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
import cntk as C
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class cntk_extractor(base_extractor):

    BASE_MODEL_URL = 'https://www.cntk.ai/Models/CNTK_Pretrained/'

    architecture_map = {
        'alexnet'              : BASE_MODEL_URL + 'AlexNet_ImageNet_CNTK.model',
        'inception_v3'         : BASE_MODEL_URL + 'InceptionV3_ImageNet_CNTK.model',
        'resnet18'             : BASE_MODEL_URL + 'ResNet18_ImageNet_CNTK.model',
        'resnet50'             : BASE_MODEL_URL + 'ResNet50_ImageNet_CNTK.model',
        'resnet101'            : BASE_MODEL_URL + 'ResNet101_ImageNet_CNTK.model',
        'resnet152'            : BASE_MODEL_URL + 'ResNet152_ImageNet_CNTK.model',
        'Fast-RCNN_grocery100' : 'https://www.cntk.ai/Models/FRCN_Grocery/Fast-RCNN_grocery100.model',
        'Fast-RCNN_Pascal'     : 'https://www.cntk.ai/Models/FRCN_Pascal/Fast-RCNN.model'
    }


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            architecture_file = download_file(cls.architecture_map[architecture], directory=path)
            model = C.Function.load(architecture_file)
            if len(model.outputs) > 1:
                for idx, output in enumerate(model.outputs):
                    if len(output.shape) > 0:
                        eval_node = idx
                        break

                model = C.as_composite(model[eval_node].owner)
                model.save(architecture_file)
                print("Cntk Model {} saved as [{}].".format(architecture, architecture_file))
            return architecture_file

        else:
            return None


    @classmethod
    def inference(cls, architecture_name, architecture_path, image_path):
        if cls.sanity_check(architecture_name):
            import numpy as np
            func = TestKit.preprocess_func['cntk'][architecture_name]
            img = func(image_path)
            img = np.transpose(img, (2, 0, 1))
            model = C.Function.load(architecture_path)
            predict = model.eval({model.arguments[0]:[img]})
            predict = np.squeeze(predict)

            top_indices = predict.argsort()[-5:][::-1]
            return predict

        else:
            return None



import sys as _sys
from six import text_type as _text_type
import sys
import imp
import os.path


def dump_code(framework, network_filepath, weight_filepath, dump_filepath, dump_tag):
    if network_filepath.endswith('.py'):
        network_filepath = network_filepath[:-3]
    sys.path.insert(0, os.path.dirname(os.path.abspath(network_filepath)))
    MainModel = imp.load_source('MainModel', network_filepath + '.py')
    if framework == 'caffe':
        from mmdnn.conversion.caffe.saver import save_model
    elif framework == 'cntk':
        from mmdnn.conversion.cntk.saver import save_model
    elif framework == 'keras':
        from mmdnn.conversion.keras.saver import save_model
    elif framework == 'mxnet':
        from mmdnn.conversion.mxnet.saver import save_model
    elif framework == 'pytorch':
        from mmdnn.conversion.pytorch.saver import save_model
    elif framework == 'tensorflow':
        from mmdnn.conversion.tensorflow.saver import save_model
        save_model(MainModel, network_filepath, weight_filepath, dump_filepath, dump_tag)
        return 0

    elif framework == 'onnx':
        from mmdnn.conversion.onnx.saver import save_model
    else:
        raise NotImplementedError("{} saver is not finished yet.".format(framework))
    save_model(MainModel, network_filepath, weight_filepath, dump_filepath)

    return 0


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Dump the model code into target model.')

    parser.add_argument(
        '-f', '--framework', type=_text_type, choices=["caffe", "cntk", "mxnet", "keras", "tensorflow", 'pytorch', 'onnx'],
        required=True,
        help='Format of model at srcModelPath (default is to auto-detect).'
    )

    parser.add_argument(
        '-in', '--inputNetwork',
        type=_text_type,
        required=True,
        help='Path to the model network architecture file.')

    parser.add_argument(
        '-iw', '--inputWeight',
        type=_text_type,
        required=True,
        help='Path to the model network weight file.')

    parser.add_argument(
        '-o', '-om', '--outputModel',
        type=_text_type,
        required=True,
        help='Path to save the target model')

    parser.add_argument(
        '--dump_tag',
        type=_text_type,
        default=None,
        help='Tensorflow model dump type',
        choices=['SERVING', 'TRAINING'])

    return parser


def _main():
    parser = _get_parser()
    args = parser.parse_args()
    ret = dump_code(args.framework, args.inputNetwork, args.inputWeight, args.outputModel, args.dump_tag)
    _sys.exit(int(ret))


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------
import os
import keras as _keras
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph


class Keras2GraphNode(GraphNode):

    def __init__(self, layer):
        super(Keras2GraphNode, self).__init__(layer)


    @property
    def name(self):
        return self.layer.name


    @property
    def type(self):
        return self.layer.__class__.__name__


    @property
    def keras_layer(self):
        return self.layer



class Keras2Graph(Graph):

    def __init__(self, model):
        # sanity check.
        if not (type(model) == _keras.models.Sequential or type(model) == _keras.models.Model):
            raise TypeError("Keras layer of type %s is not supported." % type(model))
        super(Keras2Graph, self).__init__(model)
        self.model = model


    def build(self):
        self.input_layers = list()
        for i, layer in enumerate(self.model.layers):
            self.layer_map[layer.name] = Keras2GraphNode(layer)
            self.layer_name_map[layer.name] = layer.name
            for node in layer._inbound_nodes:
                for pred in node.inbound_layers:
                    if pred.name not in self.layer_map:
                        self.layer_map[pred.name] = Keras2GraphNode(pred)
                        self.layer_name_map[pred.name] = pred.name
                    self._make_connection(pred.name, layer.name)

        # Kit: TODO
        # Duplicate models for weight sharing
        # Expand the sub-models
        super(Keras2Graph, self).build()


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from coremltools.models import datatypes

def _infer_coreml_input_shape(IR_shape, if_convert=True):
    """Infer CoreML input shape from IR shape.
    """
    if len(IR_shape) == 0:
        # the end of the tensorflow_resnet_v2_50's squeeze shape is [unknown_rank: true] with len 0
        # 1001 means the 1001 classes for tensorflow_resnet_v2_50
        # !Alert! TODO
        # Future implement can be changed to the last two layer
        shape = [1001,1,1]
    elif len(IR_shape) == 1:
        # TODO - remove style transfer 1D hack
        # Input is 1D but it goes to the width dimension: (1,1,W)
        shape = [1, 1, IR_shape[0]]  #(C,H,W)
    elif len(IR_shape) == 2:
        # assume (Batch, Channels) - Batch dimension should be dropped
        shape = [IR_shape[1]]
    elif len(IR_shape) == 3:
        # assume (Batch, Sequence-Length, channels)
        shape = [IR_shape[2], 1, IR_shape[1]]
    elif len(IR_shape) == 4:   #(B,H,W,C) --> (C,H,W)
        shape = [IR_shape[3], IR_shape[1], IR_shape[2]] #(C,H,W)
    else:
        raise ValueError('Unrecognized IR input shape {}'.format(shape))
    if if_convert:
        shape = datatypes.Array(*shape)
    return shape



from mmdnn.conversion.common.IR.IR_graph import *
import sys
import re
import numpy as np
import collections


class Folder(object):
    """A floder to fold a graph' nodes which has same scope into one node."""

    def __init__(self, graph, weights_dict, init_level=0, fold_level_num=0, scope_names=None):
        """
        Initializes a Folder.

        Args:
          graph: the graph to be folded.
          init_level: the start scope level to be folded.
          fold_level_num: the number of level from init_level to be folded. For example,
            there are three nodes, whose scope are A/B/X, A/B/Y, A/C/Z. If init_level=0 and fold_level_num=1,
            then the fold result is A/B and A/C.
        """
        self._graph = graph
        self._weights_dict = weights_dict
        self._init_level = init_level
        self._fold_level_num = fold_level_num
        self._scope_names = scope_names


    """fold the graph by compressing the nodes which have same scope into one scope node."""
    
    def fold(self):
        self.scope_level_name_map = self._get_scope_level_name_dict()  # level: scope_name set

        if self._scope_names:
            scope_names = self._scope_names
        else:
            if not self.scope_level_name_map:
                return
            scope_names = self.scope_level_name_map[0]

        for scope_name in scope_names:
            level = self._init_level
            sub_fold_level = self._fold_level_num
            while sub_fold_level >= 0:
                self._fold(self._graph.topological_sort,
                           scope_name, level, level + sub_fold_level)
                sub_fold_level -= 1

        # check the same pattern scope node whether have same inputs, outputs and weights. 
        # For those don't have, rename their scope names.
        self.check_scope()
        self.check_weights()

        # clear out scope node, typically input constant node.
        self._graph.clear_out_scope_node()


    '''
    fold the given node list by squash the nodes whose scope is given scope into one scope node.
    Args:
      scope_list: scope node topology list
      scope_name: the scope name need to be folded
      level: the scope's level
      fold_level: the scope's sub scope level need to be folded.
    '''
    
    def _fold(self, scope_list, scope_name, level, sub_level):
        top_node = None
        # get sub_scopes
        if not self.scope_level_name_map.get(sub_level, 0):
            raise ValueError("The fold level exceed maxium scope level.")
        sub_scopes = self.scope_level_name_map[sub_level]

        for sub_scope in sub_scopes:
            sub_scope_node_name_list = self._get_scope_name_dict(self._graph.topological_sort, top_level=level,
                                                                 top_scope=scope_name, sub_level=sub_level, sub_scope=sub_scope)
            for scope_list in sub_scope_node_name_list.values():
                top_node = self._graph.get_node(scope_list[-1])
                self._create_scope_node(
                    sub_scope, scope_list, top_node)


    '''get scope_level_name_dict.  {level: scope name list(without suffix number)}'''
    
    def _get_scope_level_name_dict(self):
        scope_level_name = collections.OrderedDict()
        for node in self._graph.get_nodes():
            if not node.get_attr('scope'):
                continue
            for i, scope_name in enumerate(node.get_attr('scope').split('/')):
                # decline the suffix number
                import re
                if re.search(r'\d', scope_name.split('_')[-1]):
                    scope_name = '_'.join(scope_name.split('_')[:-1])
                if scope_name == 'top':
                    continue
                if scope_level_name.get(i, None):
                    if scope_name not in scope_level_name[i]:
                        scope_level_name[i].append(scope_name)
                else:
                    scope_level_name[i] = list([scope_name]) # use list to keep sort.

        return scope_level_name


    '''
    get the dict from required node_list by appointed scope_name and level

    Args:
    node_list: current self.topology_sort
    scope_name_dict: scope_no: node_name, a dict like {scope_name_no: a set of scope node' name}
    '''
    def _get_scope_name_dict(self, node_list, top_level=0, top_scope=None, sub_level=2, sub_scope=None):
        scope_node_names = collections.OrderedDict()

        def _insert_scope_node_names_dict(scope_no, node_name):
            if scope_node_names.get(scope_no, None):
                scope_node_names[scope_no].append(node_name)
            else:
                scope_node_names[scope_no] = list([node_name])

        def _get_scope_name_dict_by_cond(cond_top, cond_sub):
            for node_name in node_list:
                node = self._graph.get_node(node_name)
                if not node.get_attr('scope'):
                    continue
                node_scope = node.get_attr('scope')
                if cond_top(top_scope, node_scope) and cond_sub(sub_scope, node_scope):
                    if 'True' in cond_top.__name__ and 'True' not in cond_sub.__name__:
                        scope_no = node_scope.split('/')[sub_level]
                    elif 'True' not in cond_top.__name__ and 'True' in cond_sub.__name__:
                        scope_no = node_scope.split('/')[top_level]
                    else:  # both not equal True
                        scope_no = node_scope.split(
                            '/')[top_level] + '_' + node_scope.split('/')[sub_level]
                    _insert_scope_node_names_dict(scope_no, node.name)

        def cond_x_in_y(x, y): return x in y

        def cond_True(x, y): return True

        # Obtain nodes where the scope name that satisfies top_level is top_scope and sub_level is sub_scope
        if top_scope and sub_scope:
            _get_scope_name_dict_by_cond(cond_x_in_y, cond_x_in_y)
        # Obtain nodes where the scope name that satisfies in sub_level is sub_scope
        elif not top_scope and sub_scope:
            _get_scope_name_dict_by_cond(cond_True, cond_x_in_y)
        # Obtain nodes where the scope name that satisfies in top_level is top_scope
        elif top_scope and not sub_scope:
            _get_scope_name_dict_by_cond(cond_x_in_y, cond_True)
        # Obtain all nodes grouped by sub_level sub_scope
        elif top_scope is None and sub_scope is None:
            top_scopes = self.scope_level_name_map[top_level]
            for top_scope in top_scopes:  # this top_scope will replace the input top_scope
                _get_scope_name_dict_by_cond(cond_x_in_y, cond_True)

        return scope_node_names


    '''get the node names' topology sort of scope nodes'''
    
    def _get_scope_nodes_topology_list(self, scope_node_name_set):

        temp_dict = {}
        for index, name in enumerate(scope_node_name_set):
            # cover the node
            self._graph.get_node(name).covered = True
            # store idx, node into a dict and sort it later to keep its topology sort.
            index = self._graph.topological_sort.index(name)
            temp_dict[name] = index

        temp_dict = sorted(
            temp_dict.items(), key=lambda item: item[1])

        return [x[0] for x in temp_dict]


    ''' rebuild the conncetion of the edge around this scope node.'''
    
    def _rebuild_scope_edges_and_get_ret_vars(self, scope_node):
        
        def _get_index(node ,name):
            for idx, in_edge in enumerate(node.in_edges):
                if in_edge.split(':')[0] == name:
                    return idx

        return_nodes = list()
        return_variable_names = list()

        for n_name in scope_node.topology_list:
            n = self._graph.get_node(n_name)
            for in_edge in n.in_edges:

                if not in_edge.split(':')[0] in scope_node.topology_list:
                    if not in_edge in scope_node.in_edges:
                        scope_node.in_edges.append(in_edge)

                    # in_node's out edges replace n_name with scope node name.
                    in_node = self._graph.get_node(in_edge)
                    if n_name in in_node.out_edges:
                        idx = in_node.out_edges.index(n_name)
                        in_node.out_edges.remove(n_name)
                        if scope_node.name not in in_node.out_edges:
                            in_node.out_edges.insert(idx, scope_node.name)

            for out_edge in n.out_edges:

                if not out_edge in scope_node.topology_list:
                    out_node = self._graph.get_node(out_edge)
                    parent_node_variable_name = self._graph.get_parent_variable_name(out_edge.split(
                        ':')[0], [_get_index(self._graph.get_node(out_edge), n_name)])

                    if parent_node_variable_name not in return_variable_names:
                        return_nodes.append(self._graph.get_node(n_name))
                        return_variable_names.append(parent_node_variable_name)
                    scope_node.out_edges.append(out_edge)

        # no out nodes means the last node in scope nodes should be returned
        if not return_nodes:
            return_nodes.append(self._graph.get_node(
                scope_node.topology_list[-1]))
            return_variable_names.append(self._graph.get_node(
                scope_node.topology_list[-1]).real_variable_name)

        ret_idx = 0
        for ret_node, ret_variable_name in zip(return_nodes, return_variable_names):

            subscript = '' if len(ret_variable_name.split(
                '[')) == 1 else ':'+ret_variable_name.split('[')[1].split(']')[0]

            for out_name in ret_node.out_edges:
                if not out_name in scope_node.topology_list:
                    out_node = self._graph.get_node(out_name)

                    ret_name = ret_node.name + subscript
                    if ret_name in out_node.in_edges:
                        insert_pos = out_node.in_edges.index(ret_name)
                        insert_name = scope_node.name + \
                            ':{}'.format(str(ret_idx)) if len(
                                return_variable_names) > 1 else scope_node.name
                        out_node.in_edges.remove(ret_name)
                        out_node.in_edges.insert(insert_pos, insert_name)

                        # if out_node is scope node, replace the scope node's inner topology list node.
                        if out_node.type == 'Scope':
                            for n in out_node.topology_list:
                                n = self._graph.get_node(n)
                                if ret_name in n.in_edges:
                                    idx = n.in_edges.index(ret_name)
                                    n.in_edges.remove(ret_name)
                                    n.in_edges.insert(idx, insert_name)
            ret_idx += 1

        return return_variable_names


    ''' if the input params include tensor which is multi-output type(e.g. unstack), then we need 
    to check whether this scope function body uses only one of the outputs or multi outputs. if it is 
    the former, feed the selected one(e.g. unstack[0]), otherwise feed all. '''
    
    def _check_and_get_input_params(self, scope_node):

        def wipe_in_egde_idx(in_name, node):
            for idx, in_edge in enumerate(node.in_edges):
                if in_name in in_edge:
                    node.in_edges[idx] = in_edge.split(':')[0]
            node.in_edges = sorted(set(node.in_edges), key=node.in_edges.index)

        input_params = list()
        in_name_dict = collections.OrderedDict()
        for in_name in scope_node.in_edges:

            if self._graph.get_node(in_name).variable_name not in input_params:
                input_params.append(self._graph.get_node(
                    in_name).variable_name)
            if ':' not in in_name:
                continue

            if in_name_dict.get(in_name.split(':')[0], None):
                in_name_dict[in_name.split(':')[0]].add(
                    in_name.split(':')[1])
            else:
                in_name_dict[in_name.split(':')[0]] = set(
                    [in_name.split(':')[1]])

        for in_name, subscript_set in in_name_dict.items():
            # the input parameter shoule be sliced when call func.
            if len(subscript_set) == 1:

                # modify the in_edges in scope inner nodes. decline the :idx.
                for n in scope_node.topology_list:
                    n = self._graph.get_node(n)
                    wipe_in_egde_idx(in_name, n)
            else:  # >2
                wipe_in_egde_idx(in_name, scope_node)

        return input_params


    '''
    create a scope node.

    Args:
    scope_name: the name of this scope, will be assigned to scope pattern.
    scope_node_names: node names involved by this scope. 
    top_node: the top node in these scope nodes.
    '''
    
    def _create_scope_node(self, scope_name, scope_node_names, top_node):
        # 1. initilize scope node
        scope_node = self._initialize_scope_node(top_node)

        # 2. get scope nodes' topology list.
        scope_node.topology_list = self._get_scope_nodes_topology_list(
            scope_node_names)
        scope_node.pattern = scope_name

        # 3. rebuild the edges connection after folding these scope nodes into one node and 
        # get this scope node's return variables.
        scope_node.return_variables = self._rebuild_scope_edges_and_get_ret_vars(
            scope_node)

        # 4. rebuild graph.
        self._graph.layer_map[scope_node.name] = scope_node
        self._graph.layer_name_map[scope_node.name] = scope_node.name
        self._graph.rebuild()


    '''initialize a scope node by copying source_node's attrs.'''
    
    def _initialize_scope_node(self, source_node):
        scope_node = self._graph.model.node.add()
        scope_node.name = source_node.name + '_scope'
        scope_node.op = 'Scope'
        scope_node = IRGraphNode(scope_node)

        kwargs = {}
        kwargs['scope'] = source_node.get_attr('scope')

        if 'data_format' in source_node.layer.attr:
            kwargs['data_format'] = source_node.get_attr('data_format')

        if '_output_shapes' in source_node.layer.attr:
            scope_node.layer.attr["_output_shapes"].MergeFromString(
                source_node.layer.attr['_output_shapes'].SerializeToString())
        if 'value' in source_node.layer.attr:
            kwargs['value'] = source_node.get_attr('value')
        # RNN-related attrs.
        if 'input_size' in source_node.layer.attr:
            kwargs['input_size'] = source_node.get_attr('input_size')
        if 'num_units' in source_node.layer.attr:
            kwargs['num_units'] = source_node.get_attr('num_units')
        if 'fill_size' in source_node.layer.attr:
            kwargs['fill_size'] = source_node.get_attr('fill_size')
        if 'fill_value' in source_node.layer.attr:
            kwargs['fill_value'] = source_node.get_attr('fill_value')

        assign_IRnode_values(scope_node.layer, kwargs)
        return scope_node


    '''
    check whether same pattern scope node has same inputs and outputs.
    For thoese do not have, rename it's pattern by adding serial number suffix.
    '''
    
    def check_scope(self):
        name_no_dict = collections.OrderedDict()
        name_inp_out_dict = collections.OrderedDict()

        for name, ir_node in self._graph.layer_map.items():
            if ir_node.type == 'Scope':
                #get input params
                ir_node.input_params = self._check_and_get_input_params(ir_node)
                origi_pattern = re.sub(r'(_\d+)*$', '', ir_node.pattern)
                if name_inp_out_dict.get(origi_pattern, None):
                    inps_and_outs = name_inp_out_dict[origi_pattern]
                    exist_Equal = False
                    for inp_out in inps_and_outs:
                        if len(ir_node.input_params) == len(inp_out[0]) and len(ir_node.return_variables)== len(inp_out[1]):
                            exist_Equal = True
                            if inp_out[2]:
                                ir_node.pattern = ir_node.pattern + '_' + str(inp_out[2])

                    if not exist_Equal:
                        name_inp_out_dict[origi_pattern].append([ir_node.input_params, ir_node.return_variables, name_no_dict.get(origi_pattern, 1)])
                        ir_node.pattern = ir_node.pattern + '_' + str(name_no_dict.get(origi_pattern, 1))
                        name_no_dict[origi_pattern] = name_no_dict.get(origi_pattern, 1) + 1

                else:
                    name_inp_out_dict[origi_pattern] = [[ir_node.input_params, ir_node.return_variables, 0]]
                    name_no_dict[ir_node.pattern] = name_no_dict.get(origi_pattern, 0) + 1
    

    '''
    check whether same pattern scope node has same weights.
    For thoese do not have, rename it's pattern by adding serial number suffix.
    '''
    
    def check_weights(self):
        weight_related_ops = ['FullyConnected']
        pattern_weight_op = collections.OrderedDict()
        name_no_dict = collections.OrderedDict()
        pattern_weights = collections.OrderedDict()

        for ir_node_name in self._graph.topological_sort:
            ir_node = self._graph.get_node(ir_node_name)
            if ir_node.type == 'Scope':
                for inner_name in ir_node.topology_list:
                    if self._graph.get_node(inner_name).type in weight_related_ops:
                        if pattern_weight_op.get(ir_node.pattern, None):
                            if self._weights_dict[inner_name]['weights'].any() and pattern_weights.get(ir_node.pattern, None):
                                inner_weights = self._weights_dict[inner_name]['weights']
                                isExist = False
                                for idx, it in enumerate(pattern_weights[ir_node.pattern]):                                        
                                    if np.array_equal(inner_weights, it['weights']):
                                        ir_node.pattern = ir_node.pattern + '_'+ str(idx)
                                        isExist = True
                                        break
                                if isExist:
                                    continue
                            pattern_weight_op[ir_node.pattern].add(inner_name)
                            pattern_weights[ir_node.pattern].append(self._weights_dict[inner_name])
                            ir_node.pattern = ir_node.pattern + '_'+ str(len(pattern_weights[ir_node.pattern])-1)

                        else:
                            pattern_weight_op[ir_node.pattern] = set([inner_name])
                            if self._weights_dict.get(inner_name, None):
                                pattern_weights[ir_node.pattern] = [self._weights_dict[inner_name]]
                                ir_node.pattern = ir_node.pattern + '_'+ str(name_no_dict.get(ir_node.pattern, 0))




#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class caffe_extractor(base_extractor):

    BASE_MODEL_URL = 'http://data.mxnet.io/models/imagenet/test/caffe/'
    MMDNN_BASE_URL = 'http://mmdnn.eastasia.cloudapp.azure.com:89/models/'

    architecture_map = {
## Image Classification
        'alexnet'       : {'prototxt'   : 'https://raw.githubusercontent.com/BVLC/caffe/master/models/bvlc_alexnet/deploy.prototxt',
                        'caffemodel' : 'http://dl.caffe.berkeleyvision.org/bvlc_alexnet.caffemodel'},
        'inception_v1'  : {'prototxt'   : 'https://raw.githubusercontent.com/BVLC/caffe/master/models/bvlc_googlenet/deploy.prototxt',
                        'caffemodel' : 'http://dl.caffe.berkeleyvision.org/bvlc_googlenet.caffemodel'},
        'vgg16'         : {'prototxt'   : 'https://gist.githubusercontent.com/ksimonyan/211839e770f7b538e2d8/raw/c3ba00e272d9f48594acef1f67e5fd12aff7a806/VGG_ILSVRC_16_layers_deploy.prototxt',
                        'caffemodel' : 'http://data.mxnet.io/models/imagenet/test/caffe/VGG_ILSVRC_16_layers.caffemodel'},
        'vgg19'         : {'prototxt'   : 'https://gist.githubusercontent.com/ksimonyan/3785162f95cd2d5fee77/raw/bb2b4fe0a9bb0669211cf3d0bc949dfdda173e9e/VGG_ILSVRC_19_layers_deploy.prototxt',
                        'caffemodel' : 'http://data.mxnet.io/models/imagenet/test/caffe/VGG_ILSVRC_19_layers.caffemodel'},
        'resnet50'      : {'prototxt'   : BASE_MODEL_URL + 'ResNet-50-deploy.prototxt',
                        'caffemodel' : BASE_MODEL_URL + 'ResNet-50-model.caffemodel'},
        'resnet101'     : {'prototxt'   : BASE_MODEL_URL + 'ResNet-101-deploy.prototxt',
                        'caffemodel' : BASE_MODEL_URL + 'ResNet-101-model.caffemodel'},
        'resnet152'     : {'prototxt'   : BASE_MODEL_URL + 'ResNet-152-deploy.prototxt',
                        'caffemodel' : BASE_MODEL_URL + 'ResNet-152-model.caffemodel'},
        'squeezenet'    : {'prototxt' : "https://raw.githubusercontent.com/DeepScale/SqueezeNet/master/SqueezeNet_v1.1/deploy.prototxt",
                           'caffemodel' : "https://github.com/DeepScale/SqueezeNet/raw/master/SqueezeNet_v1.1/squeezenet_v1.1.caffemodel"},
        'xception'      : {'prototxt' : MMDNN_BASE_URL + "caffe/xception_deploy.prototxt",
                           'caffemodel' : MMDNN_BASE_URL + "caffe/xception.caffemodel"},
        'inception_v4'  : {'prototxt' : MMDNN_BASE_URL + 'caffe/inception-v4_deploy.prototxt',
                           'caffemodel' : MMDNN_BASE_URL + 'caffe/inception-v4.caffemodel'},
## Semantic Segmentation
        'voc-fcn8s'     : {'prototxt' : 'https://raw.githubusercontent.com/shelhamer/fcn.berkeleyvision.org/master/voc-fcn8s/deploy.prototxt',
                           'caffemodel' : 'http://dl.caffe.berkeleyvision.org/fcn8s-heavy-pascal.caffemodel'},
        'voc-fcn16s'    : {'prototxt' : MMDNN_BASE_URL + "caffe/voc-fcn16s_deploy.prototxt",
                           'caffemodel' : 'http://dl.caffe.berkeleyvision.org/fcn16s-heavy-pascal.caffemodel'},
        'voc-fcn32s'    : {'prototxt' : MMDNN_BASE_URL + "caffe/voc-fcn32s_deploy.prototxt",
                           'caffemodel' : 'http://dl.caffe.berkeleyvision.org/fcn32s-heavy-pascal.caffemodel'},
    }


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            prototxt_name = architecture + "-deploy.prototxt"
            architecture_file = download_file(cls.architecture_map[architecture]['prototxt'], directory=path, local_fname=prototxt_name)
            if not architecture_file:
                return None

            weight_name = architecture + ".caffemodel"
            weight_file = download_file(cls.architecture_map[architecture]['caffemodel'], directory=path, local_fname=weight_name)
            if not weight_file:
                return None

            print("Caffe Model {} saved as [{}] and [{}].".format(architecture, architecture_file, weight_file))
            return (architecture_file, weight_file)

        else:
            return None


    @classmethod
    def inference(cls, architecture_name, architecture, path, image_path):
        if cls.sanity_check(architecture_name):
            import caffe
            import numpy as np
            net = caffe.Net(architecture[0], architecture[1], caffe.TEST)
            func = TestKit.preprocess_func['caffe'][architecture_name]
            img = func(image_path)
            img = np.transpose(img, (2, 0, 1))
            img = np.expand_dims(img, 0)
            net.blobs['data'].data[...] = img
            predict = np.squeeze(net.forward()[net._layer_names[-1]][0])
            predict = np.squeeze(predict)
            return predict

        else:
            return None



# -*- coding: utf-8 -*-
"""Normalization layers.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras.engine import Layer, InputSpec
from keras import initializers
# from keras.layers.core import Layer
from keras  import backend as K


class Scale(Layer):
    """Scale layer (2018).

    Scale the activations of the previous layer at each batch.

    # Arguments
        axis: Integer, the axis that should be normalized
            (typically the features axis).
            For instance, after a `Conv2D` layer with
            `data_format="channels_first"`,
            set `axis=1` in `BatchNormalization`.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
            When the next layer is linear (also e.g. `nn.relu`),
            this can be disabled since the scaling
            will be done by the next layer.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.

    # Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    # Output shape
        Same shape as input.

    """

    def __init__(self,
                 axis=-1,
                 center=True,
                 scale=True,
                 beta_initializer='zeros',
                 gamma_initializer='ones',
                 **kwargs):
        super(Scale, self).__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)


    def build(self, input_shape):
        dim = input_shape[self.axis]
        if dim is None:
            raise ValueError('Axis ' + str(self.axis) + ' of '
                             'input tensor should have a defined dimension '
                             'but the layer received an input with shape ' +
                             str(input_shape) + '.')
        self.input_spec = InputSpec(ndim=len(input_shape),
                                    axes={self.axis: dim})
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(shape=shape,
                                         name='gamma',
                                         initializer=self.gamma_initializer)
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(shape=shape,
                                        name='beta',
                                        initializer=self.beta_initializer)
        else:
            self.beta = None


        self.built = True

    def call(self, inputs, training=None):
        input_shape = K.int_shape(inputs)
        # Prepare broadcasting shape.
        ndim = len(input_shape)
        reduction_axes = list(range(len(input_shape)))
        del reduction_axes[self.axis]
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis]
        return K.reshape(self.gamma, broadcast_shape) * inputs + K.reshape(self.beta, broadcast_shape)

    def get_config(self):
        config = {
            'axis': self.axis,
            'center': self.center,
            'scale': self.scale,
        }
        base_config = super(Scale, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import os
from six import text_type as _text_type
from mmdnn.conversion.common.utils import download_file

BASE_MODEL_URL = 'http://data.mxnet.io/models/imagenet/test/caffe/'
# pylint: disable=line-too-long
DEFAULT_MODEL_INFO = {
    'alexnet'       : {'prototxt'   : 'https://raw.githubusercontent.com/BVLC/caffe/master/models/bvlc_alexnet/deploy.prototxt',
                       'caffemodel' : 'http://dl.caffe.berkeleyvision.org/bvlc_alexnet.caffemodel'},
    'inception_v1'  : {'prototxt'   : 'https://raw.githubusercontent.com/BVLC/caffe/master/models/bvlc_googlenet/deploy.prototxt',
                       'caffemodel' : 'http://dl.caffe.berkeleyvision.org/bvlc_googlenet.caffemodel'},
    'vgg16'         : {'prototxt'   : 'https://gist.githubusercontent.com/ksimonyan/211839e770f7b538e2d8/raw/c3ba00e272d9f48594acef1f67e5fd12aff7a806/VGG_ILSVRC_16_layers_deploy.prototxt',
                       'caffemodel' : 'http://data.mxnet.io/models/imagenet/test/caffe/VGG_ILSVRC_16_layers.caffemodel'},
    'vgg19'         : {'prototxt'   : 'https://gist.githubusercontent.com/ksimonyan/3785162f95cd2d5fee77/raw/bb2b4fe0a9bb0669211cf3d0bc949dfdda173e9e/VGG_ILSVRC_19_layers_deploy.prototxt',
                       'caffemodel' : 'http://data.mxnet.io/models/imagenet/test/caffe/VGG_ILSVRC_19_layers.caffemodel'},
    'resnet50'      : {'prototxt'   : BASE_MODEL_URL + 'ResNet-50-deploy.prototxt',
                       'caffemodel' : BASE_MODEL_URL + 'ResNet-50-model.caffemodel'},
    'resnet101'     : {'prototxt'   : BASE_MODEL_URL + 'ResNet-101-deploy.prototxt',
                       'caffemodel' : BASE_MODEL_URL + 'ResNet-101-model.caffemodel'},
    'resnet152'     : {'prototxt'   : BASE_MODEL_URL + 'ResNet-152-deploy.prototxt',
                       'caffemodel' : BASE_MODEL_URL + 'ResNet-152-model.caffemodel'},
    'squeezenet'    : {'prototxt' : "https://raw.githubusercontent.com/DeepScale/SqueezeNet/master/SqueezeNet_v1.1/deploy.prototxt",
                       'caffemodel' : "https://github.com/DeepScale/SqueezeNet/raw/master/SqueezeNet_v1.1/squeezenet_v1.1.caffemodel"}
}
# pylint: enable=line-too-long


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network', type=_text_type, help='Model Type', required=True,
                        choices=DEFAULT_MODEL_INFO.keys())

    parser.add_argument('-i', '--image', default=None,
                        type=_text_type, help='Test Image Path')

    parser.add_argument('-o', '--output_dir', default='./',
                        type=_text_type, help='Caffe Checkpoint file name')

    args = parser.parse_args()

    arch_fn = download_file(DEFAULT_MODEL_INFO[args.network]['prototxt'], directory=args.output_dir)
    if not arch_fn:
        return -1

    weight_fn = download_file(DEFAULT_MODEL_INFO[args.network]['caffemodel'], directory=args.output_dir)
    if not weight_fn:
        return -1

    print("Model {} saved.".format(args.network))

    if args.image:
        import caffe
        import numpy as np
        from mmdnn.conversion.examples.imagenet_test import TestKit

        net = caffe.Net(arch_fn.encode("utf-8"), weight_fn.encode("utf-8"), caffe.TEST)
        # net = caffe.Net(arch_fn, weight_fn, caffe.TEST)
        func = TestKit.preprocess_func['caffe'][args.network]
        img = func(args.image)
        img = np.transpose(img, (2, 0, 1))
        img = np.expand_dims(img, 0)
        net.blobs['data'].data[...] = img
        predict = np.squeeze(net.forward()['prob'][0])
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)
        print(np.sum(result))

    return 0


if __name__ == '__main__':
    _main()



from collections import namedtuple
from functools import reduce
from google.protobuf import text_format
from copy import deepcopy
import numbers
import os
import tempfile

from mmdnn.conversion.caffe.mapper import get_handler_name
from mmdnn.conversion.caffe.resolver import get_caffe_resolver, has_pycaffe
from mmdnn.conversion.caffe.shape import *
from mmdnn.conversion.caffe.errors import print_stderr, ConversionError


layer_num_to_name = {
    0: 'None',
    1: 'Accuracy',
    2: 'BNLL',
    3: 'Concat',
    4: 'Convolution',
    5: 'Data',
    6: 'Dropout',
    7: 'EuclideanLoss',
    8: 'Flatten',
    9: 'HDF5Data',
    10: 'HDF5Output',
    11: 'Im2col',
    12: 'ImageData',
    13: 'InfogainLoss',
    14: 'InnerProduct',
    15: 'LRN',
    16: 'MultinomialLogisticLoss',
    17: 'Pooling',
    18: 'ReLU',
    19: 'Sigmoid',
    20: 'Softmax',
    21: 'SoftmaxWithLoss',
    22: 'Split',
    23: 'TanH',
    24: 'WindowData',
    25: 'Eltwise',
    26: 'Power',
    27: 'SigmoidCrossEntropyLoss',
    28: 'HingeLoss',
    29: 'MemoryData',
    30: 'ArgMax',
    31: 'Threshold',
    32: 'DummyData',
    33: 'Slice',
    34: 'MVN',
    35: 'AbsVal',
    36: 'Silence',
    37: 'ContrastiveLoss',
    38: 'Exp',
    39: 'Deconvolution',
    40: 'PReLU',
    41: 'ELU',
    }

LAYER_DESCRIPTORS = {
    # Caffe Types
    'AbsVal': shape_identity,
    'Accuracy': shape_scalar,
    'ArgMax': shape_not_implemented,
    'BatchNorm': shape_identity,
    'BNLL': shape_not_implemented,
    'Concat': shape_concat,
    'ContrastiveLoss': shape_scalar,
    'Convolution': shape_convolution,
    'Crop': shape_not_implemented,
    'Deconvolution': shape_deconvolution,
    'Data': shape_data,
    'Dropout': shape_identity,
    'DummyData': shape_data,
    'EuclideanLoss': shape_scalar,
    'Eltwise': shape_identity,
    'Exp': shape_identity,
    'Flatten': shape_flatten,
    'HDF5Data': shape_data,
    'HDF5Output': shape_identity,
    'HingeLoss': shape_scalar,
    'Im2col': shape_not_implemented,
    'ImageData': shape_data,
    'InfogainLoss': shape_scalar,
    'InnerProduct': shape_inner_product,
    'Input': shape_data,
    'LRN': shape_identity,
    'MemoryData': shape_mem_data,
    'MultinomialLogisticLoss': shape_scalar,
    'MVN': shape_not_implemented,
    'Pooling': shape_pool,
    'Power': shape_identity,
    'ReLU': shape_identity,
    'Scale': shape_identity,
    'Sigmoid': shape_identity,
    'SigmoidCrossEntropyLoss': shape_scalar,
    'Silence': shape_identity,
    'Softmax': shape_identity,
    'SoftmaxWithLoss': shape_scalar,
    'Split': shape_split,
    'Slice': shape_not_implemented,
    'TanH': shape_identity,
    'WindowData': shape_not_implemented,
    'Threshold': shape_identity,
    'Reshape' : shape_reshape,
    'ResizeBilinear': shape_reshape,
    'PReLU'   : shape_identity,
    'ELU' : shape_identity,
    }

LAYER_TYPES = LAYER_DESCRIPTORS.keys()

LayerType = type('LayerType', (), {t : t for t in LAYER_TYPES})

KernelParameters = namedtuple('KernelParameters', ['global_pooling', 'k_h', 'k_w', 's_h', 's_w', 'p_h', 'p_w'])

class NodeKind(LayerType):

    @staticmethod
    def map_raw_kind(node_kind):
        if isinstance(node_kind, int):
            node_kind = layer_num_to_name[node_kind]
        else:
            node_kind = str(node_kind)
        if node_kind in LAYER_TYPES:
            return node_kind
        return None

    @staticmethod
    def compute_output_shape(node):
        try:
            return LAYER_DESCRIPTORS[node.kind](node)
        except NotImplementedError:
            raise ConversionError('Output shape computation not implemented for type: %s' % node.kind)

LAYER_IN_TRAIN_PROTO = [NodeKind.ImageData, NodeKind.Data, NodeKind.HDF5Data, NodeKind.HDF5Output, NodeKind.WindowData, NodeKind.DummyData, NodeKind.MemoryData]

class CaffeNode(object):
    def __init__(self, name, kind, layer=None):
        self.name = name
        self.kind = kind
        self.layer = layer
        self.parents = []
        self.children = []
        self.data = None
        self.output = []
        self.output_shape = None
        self.metadata = {}

    def add_parent(self, parent_node, from_output, index=None):
        assert parent_node not in self.parents
        index = len(self.parents) if index is None else index
        self.parents.insert(index, (parent_node, from_output))
        if self not in parent_node.children:
            parent_node.children.append(self)

    def get_only_parent(self):
        if len(self.parents) != 1:
            raise ConversionError('Node (%s) expected to have 1 parent. Found %s.' % (self, len(self.parents)))
        return self.parents[0]

    @property
    def parameters(self):
        if self.layer is not None:
            params = get_handler_name(self.kind)
            if params == 'deconvolution':
                params = 'convolution'
            params = '_'.join((params, 'param'))
            try:
                return getattr(self.layer, params)
            except AttributeError:
                raise ConversionError('Caffe parameters not found for layer kind: %s' % (self.kind))
        return None

    @staticmethod
    def get_kernel_value(scalar, repeated, idx, default=None):
        if scalar:
            return scalar
        if repeated:
            if isinstance(repeated, numbers.Number):
                return repeated
            if len(repeated) == 1:
                # Same value applies to all spatial dimensions
                return int(repeated[0])
            assert idx < len(repeated)
            # Extract the value for the given spatial dimension
            return repeated[idx]
        if default is None:
            raise ValueError('Unable to determine kernel parameter!')
        return default

    @property
    def kernel_parameters(self):
        assert self.kind in (NodeKind.Convolution, NodeKind.Pooling, NodeKind.Deconvolution)
        params = self.parameters
        global_pooling = hasattr(params, 'global_pooling') and params.global_pooling
        if not global_pooling:
            k_h = self.get_kernel_value(params.kernel_h, params.kernel_size, 0)
            k_w = self.get_kernel_value(params.kernel_w, params.kernel_size, 1)
            s_h = self.get_kernel_value(params.stride_h, params.stride, 0, default=1)
            s_w = self.get_kernel_value(params.stride_w, params.stride, 1, default=1)
        else:
            k_h = k_w = 0
            s_h = s_w = 1
        p_h = self.get_kernel_value(params.pad_h, params.pad, 0, default=0)
        p_w = self.get_kernel_value(params.pad_w, params.pad, 1, default=0)
        return KernelParameters(global_pooling, k_h, k_w, s_h, s_w, p_h, p_w)

    def __str__(self):
        return '[%s] %s' % (self.kind, self.name)

    def __repr__(self):
        return '%s (0x%x)' %(self.name, id(self))


class CaffeGraph(object):

    def __init__(self, nodes=None, name=None):
        self.nodes = nodes or []
        self.node_lut = {node.name: node for node in self.nodes}
        self.name = name
        self.prototxt = None

    def add_node(self, node):
        self.nodes.append(node)
        self.node_lut[node.name] = node

    def get_node(self, name):
        try:
            return self.node_lut[name]
        except KeyError:
            raise ConversionError('Layer not found: %s' % name)

    def get_input_nodes(self):
        return [node for node in self.nodes if len(node.parents) == 0]

    def get_output_nodes(self):
        return [node for node in self.nodes if len(node.children) == 0]

    def topologically_sorted(self):
        visited = set()
        sorted_nodes = []
        def topo_sort_dfs(node, visited, sorted_nodes):
            if node in visited:
                return
            visited.add(node)
            for n, idx in node.parents:
                topo_sort_dfs(n, visited, sorted_nodes)
            sorted_nodes.append(node)
        for node in self.nodes:
            topo_sort_dfs(node, visited, sorted_nodes)
        return sorted_nodes

    def compute_output_shapes(self, model):
        sorted_nodes = self.topologically_sorted()
        (tmp_handle, tmp_prototxt) = tempfile.mkstemp(suffix=".prototxt")
        with open(tmp_prototxt, 'w') as f:
            f.write(text_format.MessageToString(model))
        self.prototxt = tmp_prototxt
        if has_pycaffe():
            caffe = get_caffe_resolver().caffe
            net = caffe.Net(tmp_prototxt, caffe.TEST)
            for key, value in net.blobs.items():
                try:
                    node = self.get_node(key)
                    dims = list(value.shape)
                    dims = dims + [1] * (4 - len(dims))
                    node.output_shape = TensorShape(*dims)
                except:
                    continue
            for node in sorted_nodes:
                if node.output_shape is None:
                    node.output_shape = TensorShape(*NodeKind.compute_output_shape(node))
            os.close(tmp_handle)
        else:
            for node in sorted_nodes:
                node.output_shape = TensorShape(*NodeKind.compute_output_shape(node))

    # consider rewrite this function to Network.py
    def replaced(self, new_nodes):
        return CaffeGraph(nodes=new_nodes, name=self.name)

    def transformed(self, transformers):
        graph = self
        for transformer in transformers:
            graph = transformer(graph)
            if graph is None:
                raise ConversionError('Transformer failed: {}'.format(transformer))
            assert isinstance(graph, CaffeGraph)
        return graph

    def __contains__(self, key):
        return key in self.node_lut

    def __str__(self):
        def get_max_shape(data):
            if isinstance(data, dict):
                max = 0
                val = None
                for k, v in data.items():
                    tmp = reduce(lambda x, y: x*y, v.shape)
                    if  tmp > max:
                        val = v.shape
                        max = tmp
                return val
            else:
                return data[0].shape
        hdr = '{:<20} {:<30} {:>20} {:>20}'.format('Type', 'Name', 'Param', 'Output')
        s = [hdr, '-' * 94]
        for node in self.topologically_sorted():
            data_shape = get_max_shape(node.data) if node.data else '--'
            out_shape = node.output_shape or '--'
            s.append('{:<20} {:<30} {!s:>20} {!s:>20}'.format(node.kind, node.name, data_shape, tuple(out_shape)))
        return '\n'.join(s)


class GraphBuilder(object):
    def __init__(self, model_path, input_shape=None, is_train_proto=False, phase='test'):
        self.model_path = model_path
        self.phase = phase
        self.is_train_proto = is_train_proto
        self.input_shape = input_shape
        self.load()

    def load(self):
        self.model = get_caffe_resolver().NetParameter()
        with open(self.model_path, 'r') as f:
            text_format.Merge(f.read(), self.model)
        if self.is_train_proto:
            self.process_train_proto()

    def process_train_proto(self):
        layers = self.model.layer or self.model.layers
        delete_layer = set()
        split_op_map = dict()
        loss_layers = [layer for layer in layers if NodeKind.map_raw_kind(layer.type) in (NodeKind.SoftmaxWithLoss, NodeKind.SigmoidCrossEntropyLoss)]
        a = [layers.remove(layer) for layer in layers[:] if layer in loss_layers[:-1] or NodeKind.map_raw_kind(layer.type) in LAYER_IN_TRAIN_PROTO]
        for layer in layers[:]:
            if 'label' in layer.bottom:
                if NodeKind.map_raw_kind(layer.type) in (NodeKind.SoftmaxWithLoss, NodeKind.SigmoidCrossEntropyLoss):
                    continue
                elif NodeKind.map_raw_kind(layer.type) == NodeKind.Split:
                    for item in layer.top:
                        delete_layer.add(item)
                layers.remove(layer)
            elif NodeKind.map_raw_kind(layer.type) == NodeKind.Split:
                for item in layer.top:
                    split_op_map[item] = layer.bottom[0]
                layers.remove(layer)

        for layer in layers[:]:
            for item in delete_layer:
                if item in layer.bottom:
                    layers.remove(layer)
                    break
            for key, value in split_op_map.items():
                if key in layer.bottom:
                    layer.bottom.remove(key)
                    layer.bottom.append(value)
        self.model.input.append('data')
        self.model.input_dim.extend(self.input_shape)
        last_layer = layers[-1]
        kind = NodeKind.map_raw_kind(last_layer.type)
        if kind in (NodeKind.SoftmaxWithLoss, NodeKind.SigmoidCrossEntropyLoss):
            pred = layers.add()
            pred.name = 'prob'
            pred.top.append('prob')
            pred.bottom.append(last_layer.bottom[0])
            if kind == NodeKind.SoftmaxWithLoss:
                pred.type = NodeKind.Softmax if self.model.layer else 20 # competiable with old version caffe proto
            elif kind == NodeKind.SigmoidCrossEntropyLoss:
                pred.type = NodeKind.Sigmoid if self.model.layer else 19
        layers.remove(last_layer)

    def filter_layers(self, layers):
        phase_map = {0: 'train', 1: 'test'}
        filtered_layer_names = set()
        filtered_layers = []
        for layer in layers:
            phase = self.phase
            if len(layer.include):
                phase = phase_map[layer.include[0].phase]
            if len(layer.exclude):
                phase = phase_map[1 - layer.include[0].phase]
            exclude = (phase != self.phase)
            # Dropout layers appear in a fair number of Caffe
            # test-time networks. These are just ignored. We'll
            # filter them out here.
            if (not exclude) and (phase == 'test'):
                exclude = (layer.type == LayerType.Dropout)
            if (not exclude):
                exclude = (layer.type == LayerType.Silence)
            if not exclude:
                if layer.name in filtered_layer_names:
                    for i in range(1, len(filtered_layer_names)):
                        new_name = layer.name + '_%s' % i
                        if new_name not in filtered_layer_names:
                            layer.name = new_name
                            break
                filtered_layer_names.add(layer.name)
                filtered_layers.append(layer)
        return filtered_layers

    def make_node(self, layer):
        kind = NodeKind.map_raw_kind(layer.type)
        if kind is None:
            # TODO: raise error
            pass
        node = CaffeNode(layer.name, kind, layer=layer)
        node.output.append(layer.name.replace('/', '_'))
        node.output.extend(layer.top[1:])
        return node

    def make_input_node(self):
        nodes = [CaffeNode(name, NodeKind.Data) for name in self.model.input]
        if len(nodes):
            input_dim = list(map(int, self.model.input_dim))
            if not input_dim:
                if len(self.model.input_shape) > 0:
                    input_dim = list(map(int, self.model.input_shape[0].dim))
                else:
                    # TODO: raise error
                    pass
            for node in nodes:
                node.output_shape = tuple(input_dim)
                node.output.append('data')
        return nodes

    def build(self):
        layers = self.model.layers or self.model.layer
        layers = self.filter_layers(layers)
        nodes = self.make_input_node()
        nodes += [self.make_node(layer) for layer in layers]
        graph = CaffeGraph(nodes=nodes, name=self.model.name)
        node_outputs = {}
        for idx, layer in enumerate(layers):
            node = graph.get_node(layer.name)
            for input_name in layer.bottom:
                assert input_name != layer.name
                parent_node = node_outputs.get(input_name)
                if (parent_node is None) or (parent_node==node):
                    parent_node = graph.get_node(input_name)
                if parent_node.layer:
                    for i, output in enumerate(parent_node.layer.top):
                        if input_name == output:
                            node.add_parent(parent_node, i)
                else:
                    node.add_parent(parent_node, 0)
            for output_name in layer.top:
                if output_name == layer.name:
                    continue
                node_outputs[output_name] = node
        graph.compute_output_shapes(self.model)
        return graph



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import cntk as _cntk
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph


class CntkGraphNode(GraphNode):

    def __init__(self, layer):
        super(CntkGraphNode, self).__init__(layer)


    @property
    def name(self):
        return self.layer.uid


    @property
    def type(self):
        if hasattr(self.layer, 'op_name'):
            return self.layer.op_name
        elif self.layer.is_input:
            return "DataInput"
        else:
            raise NotImplementedError()


    @property
    def cntk_layer(self):
        return self.layer


    def get_attr(self, name, default_value=None):
        if self.layer.is_block:
            return self.layer.block_root.attributes[name]
        else:
            return self.layer.attributes[name]


class CntkGraph(Graph):

    def __init__(self, model):
        # sanity check.
        pass

        self.weights = dict()
        self.visited = set()
        super(CntkGraph, self).__init__(model)


    def _traverse_graph(self, son_node):
        if not son_node.uid in self.visited:
            self.visited.add(son_node.uid)

            if son_node.is_block:
                inputs = [input for _, input in son_node.block_arguments_mapping]

            else:
                inputs = son_node.inputs

            for input_node in inputs:
                if input_node.is_output:
                    input_node = input_node.owner
                    if not input_node.uid in self.layer_map:
                        self.layer_map[input_node.uid] = CntkGraphNode(input_node)
                    self._make_connection(input_node.uid, son_node.uid)
                    self._traverse_graph(input_node)

                elif input_node.is_input:
                    if not input_node.uid in self.layer_map:
                        self.layer_map[input_node.uid] = CntkGraphNode(input_node)
                    self._make_connection(input_node.uid, son_node.uid)

                elif input_node.is_placeholder:
                    raise NotImplementedError("PlaceHolder of placeholder is not supported.")


    def build(self):
        if len(self.model.outputs) > 1:
            for idx, output in enumerate(self.model.outputs):
                if len(output.shape) > 0:
                    eval_node = idx
                    break

            output = self.model[eval_node].owner
        else:
            output = self.model.outputs[0].owner

        self.layer_map[output.uid] = CntkGraphNode(output)
        self._traverse_graph(output)

        super(CntkGraph, self).build()

"""
    def __traverse_graph(self, node):
        if node.uid in self.visited:
            return

        self.visited.add(node.uid)

        if isinstance(node, _cntk.Function) and node.is_block:
            composite = node.block_root

            # BlockFunction node
            mapping = node.block_arguments_mapping

            # redirect the composite's inputs to the true inputs
            stack.extend([(actual_input, depth-1) for _, actual_input in mapping]) # traverse into actual composite inputs
            visited |= {comp_input.uid for comp_input, _ in mapping}    # don't traverse into the mapped-away inputs
            stack.append((composite, depth-1))
            # BlockFunctions are short-circuited, and not added to accum[]
        try:
            # Function node
            stack = list((i, depth) for i in node.root_function.inputs) + stack
        except AttributeError:
            # OutputVariable node
            try:
                if node.is_output:
                    stack.insert(0, (node.owner, depth))
                    visited.add(node.uid)
                    continue
            except AttributeError:
                pass

        if visitor(node):
            if isinstance(node, Variable):
                if node.is_parameter:
                    node = node.as_parameter()
                elif node.is_constant:
                    node = node.as_constant()

            accum.append(node)

        visited.add(node.uid)


    # def build(self):
    #     _traverse_graph(self, self.model.root_function)
    #     super(CntkGraph, self).build()
"""